{%- if not pagename in ("search", "404", "genindex") and hasdoc(pagename) %}
<div class="footer-contribute">
  <ul>
    <li>
      <a href="https://projects.blender.org/blender/blender/issues/new?{#
               #}template=.gitea/issue_template/api_docs.yaml&field:{#
               #}body=%2A%2APage+Information%2A%2A%0D%0A{#
               #}File%3A+%60{{ pagename }}{{ page_source_suffix }}%60%0D%0A{#
               #}Blender+Version%3A+%60{{ version }}%60%0D%0A%5B{#
               #}Permanent+Link%5D%28https%3A%2F%2Fdocs.blender.org%2F{#
               #}api%2F{{ version }}%2F{{ pagename }}{{ file_suffix }}%29%0D%0A%0D%0A%2A%2A{#
               #}Short+description+of+error%2A%2A%0D%0A%5B{#
               #}Please+fill+out+a+short+description+of+the+error+here%5D%0D%0A" class="fa fa-bug"> {{ _('Report issue
        on this page') }}</a>
    </li>
  </ul>
</div>
{%- endif %}
