/* SPDX-FileCopyrightText: 2001-2002 NaN Holding BV. All rights reserved.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

void *memdbl(void *mem, int *size_pr, int item_size);

char *string_dup(char *str);
void fatal(char *fmt, ...);
