/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

#pragma once

struct ARegionType;

namespace blender::ed::spreadsheet {

void register_row_filter_panels(ARegionType &region_type);

}
