/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

/** \file
 * \ingroup spoutliner
 */

#pragma once

#include "tree_element_id.hh"

struct bGPdata;

namespace blender::ed::outliner {

class TreeElementIDGPLegacy final : public TreeElementID {
  bGPdata &gpd_;

 public:
  TreeElementIDGPLegacy(TreeElement &legacy_te, bGPdata &gpd);

  void expand(SpaceOutliner & /*soops*/) const override;

 private:
  void expand_layers() const;
};

}  // namespace blender::ed::outliner
