/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

#include "spreadsheet_data_source.hh"

namespace blender::ed::spreadsheet {

/* Provide a "key function" for the linker. */
DataSource::~DataSource() = default;

}  // namespace blender::ed::spreadsheet
