/* SPDX-FileCopyrightText: 2011-2022 Blender Foundation
 *
 * SPDX-License-Identifier: Apache-2.0 */

#pragma once

#include <memory>

CCL_NAMESPACE_BEGIN

using std::make_unique;
using std::unique_ptr;

CCL_NAMESPACE_END
