/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

#pragma once

struct wmOperatorType;

void WM_OT_drop_import_file(wmOperatorType *ot);
void ED_dropbox_drop_import_file();
