/* SPDX-FileCopyrightText: 2001-2002 NaN Holding BV. All rights reserved.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

#pragma once

#include "DNA_userdef_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/** Default theme, see: `release/datafiles/userdef/userdef_default_theme.c`. */
extern const bTheme U_theme_default;
/** Default preferences, defined by: `release/datafiles/userdef/userdef_default.c`. */
extern const UserDef U_default;

#ifdef __cplusplus
}
#endif
