/* SPDX-FileCopyrightText: 2025 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

#pragma once
/* This file must replaced at runtime. The following content is only a possible implementation. */
#pragma runtime_generated

/* Functions generated from #ShaderOperation::generate_code_for_outputs. */
/* Commented out because they mess with the library parser. */
/**
 * void store_float(const uint id, float value) {}
 * void store_int(const uint id, float value) {}
 * void store_bool(const uint id, float value) {}
 * void store_float3(const uint id, vec3 value) {}
 * void store_color(const uint id, vec4 value) {}
 * void store_float4(const uint id, vec4 value) {}
 * void store_float2(const uint id, vec3 value) {}
 * void store_int2(const uint id, vec3 value) {}
 * void store_menu(const uint id, float value) {}
 */
