/* SPDX-FileCopyrightText: 2016 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

/** \file
 * \ingroup draw
 */

#pragma once

void DRW_draw_region_info(const bContext *C, ARegion *region);
void DRW_draw_gizmo_3d(const bContext *C, ARegion *region);
void DRW_draw_gizmo_2d(const bContext *C, ARegion *region);
