/* SPDX-FileCopyrightText: 2018 Blender Foundation
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Author: Sergey Sharybin. */

#include "opensubdiv_capi.hh"

#include <cstddef>

void openSubdiv_init() {}

void openSubdiv_cleanup() {}

int openSubdiv_getVersionHex()
{
  return 0;
}
