/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

void main()
{
  int index = int(gl_GlobalInvocationID.x);
  float value = texelFetch(bufferTexture, index).r;
  data_out[index] = value;
}
