/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

/** \file
 * \ingroup bli
 */

#define KD_DIMS 4
#define KDTREE_PREFIX_ID BLI_kdtree_4d
#define KDTree KDTree_4d
#define KDTreeNode KDTreeNode_4d
#define KDTreeNode_head KDTreeNode_head_4d
#define KDTreeNearest KDTreeNearest_4d
#include "kdtree_impl.h"
