/* SPDX-FileCopyrightText: 2021-2022 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0 */
}
; /* end of ONEAPIKernelContext class definition */

#undef kernel_integrator_state
#define kernel_integrator_state (*(kg->integrator_state))
