/* SPDX-FileCopyrightText: 2024 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

/** \file
 * \ingroup bli
 */

#pragma once

namespace blender::index_mask {

class IndexMask;
class IndexMaskMemory;

}  // namespace blender::index_mask

namespace blender {

using index_mask::IndexMask;
using index_mask::IndexMaskMemory;

}  // namespace blender
