/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

/** \file
 * \ingroup DNA
 */

#pragma once

/* clang-format off */

/* -------------------------------------------------------------------- */
/** \name Curves Struct
 * \{ */

#define _DNA_DEFAULT_Curves \
  { \
    .flag = 0, \
    .surface_collision_distance = 0.005f, \
  }

/** \} */

/* clang-format on */
